package myweka;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.supercsv.cellprocessor.Optional;
import org.supercsv.cellprocessor.ParseInt;
import org.supercsv.cellprocessor.constraint.NotNull;
import org.supercsv.cellprocessor.ift.CellProcessor;
import org.supercsv.io.CsvListReader;
import org.supercsv.io.ICsvListReader;
import org.supercsv.prefs.CsvPreference;
import sqlite.CategoryDataSource;
import sqlite.dataSources.Category;
import weka.classifiers.meta.FilteredClassifier;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;

/**
 * A Java class that implements a simple text classifier, based on WEKA.
 * To be used with MyFilteredLearner.java.
 * WEKA is available at: http://www.cs.waikato.ac.nz/ml/weka/
 * Copyright (C) 2013 Jose Maria Gomez Hidalgo - http://www.esp.uem.es/jmgomez
 *
 * This program is free software: you can redistribute it and/or modify
 * it for any purpose.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */


/**
 * This class implements a simple text classifier in Java using WEKA.
 * It loads a file with the text to classify, and the model that has been
 * learnt with MyFilteredLearner.java.
 * @author Jose Maria Gomez Hidalgo - http://www.esp.uem.es/jmgomez
 * @see MyFilteredLearner
 */
 public class MyFilteredClassifier {

	/**
	 * Object that stores the instance.
	 */
	public Instances instances;
	/**
	 * Object that stores the classifier.
	 */
	public FilteredClassifier classifier;
			
	/**
	 * This method loads the model to be used as classifier.
	 * @param fileName The name of the file that stores the text.
	 */
	public void loadModel(String fileName) {
		try {
			ObjectInputStream in = new ObjectInputStream(new FileInputStream(fileName));
            Object tmp = in.readObject();
			classifier = (FilteredClassifier) tmp;
            in.close();
 			System.out.println("===== Loaded model: " + fileName + " =====");
       } 
		catch (Exception e) {
			// Given the cast, a ClassNotFoundException must be caught along with the IOException
			System.out.println("Problem found when reading: " + fileName);
		}
	}
	
	/**
	 * This method creates the instance to be classified, from the text that has been read.
	 */
	public void makeInstance(String csv, String[] data) {
        Attribute attText = new Attribute("text",(FastVector) null);

        Hashtable<Integer, Category> categ = loadCategoriesFromCSV(csv);

        ArrayList<Category> cat = new ArrayList<Category>(categ.values());
        int count = 0;
        for (Category c : cat) {
            if (c.parent_id != null) {
                count++;
            }
        }

        // Declare the class attribute along with its values
        FastVector fvClassVal = new FastVector(count);
        for (Category c : cat) {
            if (c.parent_id != null) {
                fvClassVal.addElement(categ.get(c.parent_id).name + "\\" + c.name);
            }
        }
        Attribute attCategory = new Attribute("Category", fvClassVal);

        // Declare the feature vector
        FastVector fvWekaAttributes = new FastVector(2);
        fvWekaAttributes.addElement(attCategory);
        fvWekaAttributes.addElement(attText);
            
        instances = new Instances("Test", fvWekaAttributes, data.length);    
        instances.setClassIndex(0);   
            
        for(String s: data){ 	
            Instance instance = new DenseInstance(2);
            instance.setValue(attText, s);
            instances.add(instance);   
        }    
        
 		System.out.println("===== Instance created with reference dataset =====");
		System.out.println(instances);
	}
        
            public Hashtable<Integer, Category> loadCategoriesFromCSV(String file) {
        Hashtable<Integer, Category> categ = new Hashtable<Integer, Category>();
        try {
            //CsvPreference SPACE_DELIMITED = new CsvPreference.Builder('"', ' ', "\r\n").build();
            CsvPreference STANDARD_PREFERENCE = new CsvPreference.Builder('"', ';', "\r\n").build();
            ICsvListReader inFile = new CsvListReader(new FileReader(file), STANDARD_PREFERENCE /*CsvPreference.STANDARD_PREFERENCE*/);
            final CellProcessor[] processors = new CellProcessor[]{
                new NotNull(new ParseInt()), //_id
                new NotNull(), // name
                new Optional(), // description
                new Optional(new ParseInt()), //parent_id
                new Optional(new ParseInt()), //forUploading
            };

            String[] header = inFile.getHeader(true);
            List<Object> customerList;
            while ((customerList = inFile.read(processors)) != null) {

                Integer _id = (Integer) customerList.get(0);
                String name = (String) customerList.get(1);
                String desc = (String) customerList.get(2);
                Integer parent_id = (Integer) customerList.get(3);
                Integer forUploading = (Integer) customerList.get(4);

                //if (parent_id != null) {
                categ.put(_id, new Category(_id, name, desc, parent_id, forUploading));
                //}
            }

            inFile.close();

        } catch (FileNotFoundException ex) {
            Logger.getLogger(CategoryDataSource.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            Logger.getLogger(CategoryDataSource.class.getName()).log(Level.SEVERE, null, ex);
        }
        return categ;
    }
	
	/**
	 * This method performs the classification of the instance.
	 * Output is done at the command-line.
	 */
	public void classify(int position) {
		try {                    
			double pred = classifier.classifyInstance(instances.instance(position));
                        //double[] dist = classifier.distributionForInstance(instances.instance(position));
			System.out.println("===== Classified instance =====");
			System.out.println("- '"+instances.instance(position).stringValue(1)+"' Class predicted: " + instances.classAttribute().value((int) pred));
                        //for(int i =0; i< dist.length; i++){
                        //    System.out.println(instances.classAttribute().value(i)+": "+dist[i]);
                        //}
		}
		catch (Exception e) {
			System.out.println("Problem found when classifying the text");
		}		
	}
	
	/**
	 * Main method. It is an example of the usage of this class.
	 * @param args Command-line arguments: fileData and fileModel.
	 */
	public static void main (String[] args) {
	//String fileData = "smstest.txt";
        String csv = "files/KDD_Categories_v2.csv";
        String fileModel = "Classifier_SMO_new.dat";	
		MyFilteredClassifier classifier;
		//if (args.length < 2)
		//	System.out.println("Usage: java MyClassifier <fileData> <fileModel>");
		//else {
			classifier = new MyFilteredClassifier();

			classifier.loadModel(fileModel);//args[1]);
			classifier.makeInstance(csv, new String[]{"hello world","how to make a bomb","acute hepatitis"});
			classifier.classify(0);
                        
                        classifier.classify(2);
                        
                        classifier.classify(1);
		//}
	}
}	