package myweka;

/**
 * A Java class that implements a simple text learner, based on WEKA. To be used
 * with MyFilteredClassifier.java. WEKA is available at:
 * http://www.cs.waikato.ac.nz/ml/weka/ Copyright (C) 2013 Jose Maria Gomez
 * Hidalgo - http://www.esp.uem.es/jmgomez
 *
 * This program is free software: you can redistribute it and/or modify it for
 * any purpose.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 */
import weka.core.Instances;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.StringToWordVector;
import weka.classifiers.Evaluation;
import java.util.Random;
import weka.classifiers.bayes.NaiveBayes;
import weka.classifiers.meta.FilteredClassifier;
import weka.core.converters.ArffLoader.ArffReader;
import java.io.*;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import lemur.dataStructures.CategoryScoreRank;
import lemur.dataStructures.DocumentOfIndex;
import lemur.dataStructures.ListOfDocumentsInIndex;
import org.supercsv.cellprocessor.Optional;
import org.supercsv.cellprocessor.ParseInt;
import org.supercsv.cellprocessor.constraint.NotNull;
import org.supercsv.cellprocessor.ift.CellProcessor;
import org.supercsv.io.CsvListReader;
import org.supercsv.io.ICsvListReader;
import org.supercsv.prefs.CsvPreference;
import sqlite.CategoryDataSource;
import sqlite.dataSources.Category;
import weka.classifiers.Classifier;
import weka.classifiers.functions.LibSVM;
import weka.classifiers.functions.SMO;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.SelectedTag;
import weka.core.tokenizers.NGramTokenizer;

/**
 * This class implements a simple text learner in Java using WEKA. It loads a
 * text dataset written in ARFF format, evaluates a classifier on it, and saves
 * the learnt model for further use.
 *
 * @author Jose Maria Gomez Hidalgo - http://www.esp.uem.es/jmgomez
 * @see MyFilteredClassifier
 */
public class MyFilteredLearner {

    /**
     * Object that stores training data.
     */
    public Instances trainData;
    /**
     * Object that stores the filter
     */
    public StringToWordVector filter;
    /**
     * Object that stores the classifier
     */
    public FilteredClassifier classifier;

    /**
     * This method loads a dataset in ARFF format. If the file does not exist,
     * or it has a wrong format, the attribute trainData is null.
     *
     * @param fileName The name of the file that stores the dataset.
     */
    public void loadDataset(String fileName) {
        try {
            BufferedReader reader = new BufferedReader(new FileReader(fileName));
            ArffReader arff = new ArffReader(reader);
            trainData = arff.getData();
            System.out.println("===== Loaded dataset: " + fileName + " =====");
            reader.close();
        } catch (IOException e) {
            System.out.println("Problem found when reading: " + fileName);
        }
    }

    public void loadDataset(String csv, String data) {

        Attribute attText = new Attribute("text",(FastVector) null);

        Hashtable<Integer, Category> categ = loadCategoriesFromCSV(csv);

        ArrayList<Category> cat = new ArrayList<Category>(categ.values());
        int count = 0;
        for (Category c : cat) {
            if (c.parent_id != null) {
                count++;
            }
        }

        // Declare the class attribute along with its values
        FastVector fvClassVal = new FastVector(count);
        for (Category c : cat) {
            if (c.parent_id != null) {
                fvClassVal.addElement(categ.get(c.parent_id).name + "\\" + c.name);
            }
        }
        Attribute attCategory = new Attribute("Category", fvClassVal);

        // Declare the feature vector
        FastVector fvWekaAttributes = new FastVector(2);
        fvWekaAttributes.addElement(attCategory);
        fvWekaAttributes.addElement(attText);

        ListOfDocumentsInIndex lod = new ListOfDocumentsInIndex();
        lod.loadSavedObject(data);

        trainData = new Instances("Rel", fvWekaAttributes, lod.docs.size());
        //trainData.setClassIndex(0);

        Enumeration<Integer> en = lod.docs.keys();
        int total = 0;
        while (en.hasMoreElements()) {
            int docid = en.nextElement();
            DocumentOfIndex doi = lod.docs.get(docid);
            for (int r =0 ;r<doi.Categs.size();r++) {
                Instance i = new DenseInstance(2);
                i.setValue((Attribute) fvWekaAttributes.elementAt(0), categ.get(doi.Categs.get(r).parent_id).name + "\\" + doi.Categs.get(r).name);
                i.setValue((Attribute) fvWekaAttributes.elementAt(1), doi.text);
                trainData.add(i);
                total++;
            }
        }
        System.out.println("--- Total Documents for training: "+total);
    }

    public Hashtable<Integer, Category> loadCategoriesFromCSV(String file) {
        Hashtable<Integer, Category> categ = new Hashtable<Integer, Category>();
        try {
            //CsvPreference SPACE_DELIMITED = new CsvPreference.Builder('"', ' ', "\r\n").build();
            CsvPreference STANDARD_PREFERENCE = new CsvPreference.Builder('"', ';', "\r\n").build();
            ICsvListReader inFile = new CsvListReader(new FileReader(file), STANDARD_PREFERENCE /*CsvPreference.STANDARD_PREFERENCE*/);
            final CellProcessor[] processors = new CellProcessor[]{
                new NotNull(new ParseInt()), //_id
                new NotNull(), // name
                new Optional(), // description
                new Optional(new ParseInt()), //parent_id
                new Optional(new ParseInt()), //forUploading
            };

            String[] header = inFile.getHeader(true);
            List<Object> customerList;
            while ((customerList = inFile.read(processors)) != null) {

                Integer _id = (Integer) customerList.get(0);
                String name = (String) customerList.get(1);
                String desc = (String) customerList.get(2);
                Integer parent_id = (Integer) customerList.get(3);
                Integer forUploading = (Integer) customerList.get(4);

                //if (parent_id != null) {
                categ.put(_id, new Category(_id, name, desc, parent_id, forUploading));
                //}
            }

            inFile.close();

        } catch (FileNotFoundException ex) {
            Logger.getLogger(CategoryDataSource.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            Logger.getLogger(CategoryDataSource.class.getName()).log(Level.SEVERE, null, ex);
        }
        return categ;
    }

    /**
     * This method evaluates the classifier. As recommended by WEKA
     * documentation, the classifier is defined but not trained yet. Evaluation
     * of previously trained classifiers can lead to unexpected results.
     */
    public void evaluate(Classifier cl) {
        try {
            trainData.setClassIndex(0);
            //http://www.uky.edu/~nyu222/tutorials/Weka.htm
            filter = new StringToWordVector();
            
            filter.setAttributeIndices("last");
            filter.setWordsToKeep(5000);
            filter.setOutputWordCounts(true);
            filter.setTFTransform(true);
            filter.setIDFTransform(true);
            filter.setNormalizeDocLength(new SelectedTag(1, StringToWordVector.TAGS_FILTER)); //Normalize all data
            filter.setLowerCaseTokens(true);
            filter.setMinTermFreq(2);
            filter.setTokenizer(new NGramTokenizer());
            
            classifier = new FilteredClassifier();
            classifier.setFilter(filter);
            classifier.setClassifier(cl);
            Evaluation eval = new Evaluation(trainData);
            eval.crossValidateModel(classifier, trainData, 4, new Random(1));
            System.out.println(eval.toSummaryString());
            System.out.println(eval.toClassDetailsString());
            System.out.println("===== Evaluating on filtered (training) dataset done =====");
        } catch (Exception e) {
            System.out.println("Problem found when evaluating");
        }
    }

    /**
     * This method trains the classifier on the loaded dataset.
     */
    public void learn(Classifier cl) {
        try {
            trainData.setClassIndex(0);
            //http://www.uky.edu/~nyu222/tutorials/Weka.htm
            filter = new StringToWordVector();
            
            filter.setAttributeIndices("last");
            filter.setWordsToKeep(5000);
            filter.setOutputWordCounts(true);
            filter.setTFTransform(true);
            filter.setIDFTransform(true);
            filter.setNormalizeDocLength(new SelectedTag(1, StringToWordVector.TAGS_FILTER)); //Normalize all data
            filter.setLowerCaseTokens(true);
            filter.setMinTermFreq(2);
            filter.setTokenizer(new NGramTokenizer());
            
            classifier = new FilteredClassifier();
            classifier.setFilter(filter);
            classifier.setClassifier(cl);
            classifier.buildClassifier(trainData);
            // Uncomment to see the classifier
            // System.out.println(classifier);
            System.out.println("===== Training on filtered (training) dataset done =====");
        } catch (Exception e) {
            System.out.println("Problem found when training");
        }
    }

    /**
     * This method saves the trained model into a file. This is done by simple
     * serialization of the classifier object.
     *
     * @param fileName The name of the file that will store the trained model.
     */
    public void saveModel(String fileName) {
        try {
            ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(fileName));
            out.writeObject(classifier);
            out.close();
            System.out.println("===== Saved model: " + fileName + " =====");
        } catch (IOException e) {
            System.out.println("Problem found when writing: " + fileName);
        }
    }

    /**
     * Main method. It is an example of the usage of this class.
     *
     * @param args Command-line arguments: fileData and fileModel.
     */
    public static void main(String[] args) {
        String csv = "files/KDD_Categories_v2.csv";
        String data = "files/ListOfDocumentsInIndex_kddcup05_v2.data";

        //String fileData = "smsspam.small.arff";
        String fileModel = "Classifier_SMO_new_NGRAM.dat";
        MyFilteredLearner learner;
        //if (args.length < 2)
        //	System.out.println("Usage: java MyLearner <fileData> <fileModel>");
        //else {
        learner = new MyFilteredLearner();
        learner.loadDataset(csv, data);
        //learner.loadDataset(fileData);//args[0]);
        // Evaluation mus be done before training
        // More info in: http://weka.wikispaces.com/Use+WEKA+in+your+Java+code
        //learner.evaluate(new NaiveBayes());
        learner.evaluate(new SMO());
        //learner.evaluate(new LibSVM());

        //learner.learn(new NaiveBayes());
        learner.learn(new SMO());
        //learner.learn(new LibSVM());
        learner.saveModel(fileModel);//args[1]);
        //}
    }
}
